@@ -12,16 +12,17 @@ our @ISA = qw(Exporter);
 our @EXPORT_OK = qw( canon_ie format_ie parse_ie random_ie );
 our @EXPORT = qw( test_ie );
 
-our $VERSION = '0.00_14';
+our $VERSION = '0.00_18'; # ???
+$VERSION = eval $VERSION;
 
-use Business::BR::Ids::Common qw(_dot _canon_i);
+use Business::BR::Ids::Common qw( _dot _dot_10 _canon_id );
 
 ### AC ###
 
 # http://www.sintegra.gov.br/Cad_Estados/cad_AC.html
 
 sub canon_ie_ac {
-  return _canon_i(shift, size => 13);
+  return _canon_id(shift, size => 13);
 }
 sub test_ie_ac {
   my $ie = canon_ie_ac shift;
@@ -84,7 +85,7 @@ my %AL_TYPES = (
 my @AL_TYPES = keys %AL_TYPES;
 
 sub canon_ie_al {
-  return _canon_i(shift, size => 9);
+  return _canon_id(shift, size => 9);
 }
 sub test_ie_al {
   my $ie = canon_ie_al shift;
@@ -98,7 +99,7 @@ sub test_ie_al {
 
 }
 sub format_ie_al {
-  my $ie = canon_ie_ac shift;
+  my $ie = canon_ie_al shift;
   $ie =~ s|^(..)(...)(...)(.).*|$1.$2.$3-$4|; # 24.000.004-8
   return $ie;
 }
@@ -135,12 +136,116 @@ sub parse_ie_al {
   };
 }
 
+### AP ###
+
+#***
+
+sub canon_ie_ap {
+  return _canon_id(shift, size => 9);
+}
+sub test_ie_ap {
+  my $ie = canon_ie_ap shift;
+  return undef if length $ie != 9;
+  return 0 unless $ie =~ /^03/;
+
+  my $nr_empresa = substr($ie, 2, -1);
+  my ($p, $d) = ($nr_empresa >= 1) && ($nr_empresa <= 17000) ? (5, 0) : # 1st class, 03.000.001-x up to 03.017.000-x
+                ($nr_empresa >= 17001) && ($nr_empresa <= 19022) ? (9, 1) : # 2nd class, 03.017.001-x up to 03.019.022-x
+                (0, 0); # 3rd class, from 03.019.023-x and on
+#  print "(p, d) = ($p, $d)\n";
+
+  my @ie = split '', $ie;
+  my $sum = -($p + _dot([9, 8, 7, 6, 5, 4, 3, 2, 1], \@ie)) % 11; 
+
+#  print "# ie: $ie, sum: $sum\n"; # ***
+  return ($sum==0 || $sum==10 && $ie[8]==0) ? 1 : 0;
+
+} # FIXME: this is not QUITE RIGHT !!!!!!!!!
+sub format_ie_ap {
+  my $ie = canon_ie_ap shift;
+  $ie =~ s|^(..)(...)(...)(.).*|$1.$2.$3-$4|; # 03.012.245-9
+  return $ie;
+}
+sub _dv_ie_ap {
+  my $base = shift; # expected to be canon'ed already ?!
+  my $valid = @_ ? shift : 1;
+  my $dev = $valid ? 0 : 3; # deviation (to make IE-AP invalid)
+  my @base = split '', $base;
+
+  my $nr_empresa = substr($base, 2, -1);
+  my ($p, $d) = ($nr_empresa >= 1) && ($nr_empresa <= 17000) ? (5, 0) : # 1st class, 03.000.001-x up to 03.017.000-x
+                ($nr_empresa >= 17001) && ($nr_empresa <= 19022) ? (9, 1) : # 2nd class, 03.017.001-x up to 03.019.022-x
+                (0, 0); # 3rd class, from 03.019.023-x and on
+
+  my $dv1 = -($p + _dot([9, 8, 7, 6, 5, 4, 3, 2, 0], \@base) + $dev) % 11 % 10;
+  return ($dv1) if wantarray;
+  substr($base, 8, 1) = $dv1;
+  return $base;
+}
+sub random_ie_ap {
+  my $valid = @_ ? shift : 1; # valid IE-AP by default
+  my $base = sprintf "03%06d*", 
+                   int(rand(1E6)); # '03', 6 digits, dv
+  return scalar _dv_ie_ap($base, $valid);
+}
+sub parse_ie_ap {
+  my $ie = canon_ie_ap shift;
+  my ($base, $dv) = $ie =~ /(\d{8})(\d{1})/;
+  if (wantarray) {
+    return ($base, $dv);
+  }
+  return { 
+    base => $base, 
+    dv => $dv, 
+    range => '?'
+  };
+
+}
+
+
+
+### AM ###
+
+sub canon_ie_am {
+  return _canon_id(shift, size => 9);
+}
+
+### BA ###
+
+sub canon_ie_ba {
+  return _canon_id(shift, size => 8);
+}
+
+### CE ###
+
+sub canon_ie_ce {
+  return _canon_id(shift, size => 9);
+}
+
+### DF ###
+
+sub canon_ie_df {
+  return _canon_id(shift, size => 13);
+}
+
+### ES ###
+
+sub canon_ie_es {
+  return _canon_id(shift, size => 9);
+}
+
+### GO ###
+
+sub canon_ie_go {
+  return _canon_id(shift, size => 9);
+}
+
 ### MA ###
 
 # http://www.sintegra.gov.br/Cad_Estados/cad_MA.html
 
 sub canon_ie_ma {
-  return _canon_i(shift, size => 9);
+  return _canon_id(shift, size => 9);
 }
 sub test_ie_ma {
   my $ie = canon_ie_ma shift;
@@ -180,6 +285,89 @@ sub parse_ie_ma {
   return { base => $base, dv => $dv };
 }
 
+### MT ###
+
+sub canon_ie_mt {
+  return _canon_id(shift, size => 11);
+}
+
+### MS ###
+
+sub canon_ie_ms {
+  return _canon_id(shift, size => 9);
+}
+
+### MG ###
+
+# http://www.sintegra.gov.br/Cad_Estados/cad_MG.html
+
+sub canon_ie_mg {
+  return _canon_id( shift, size => 13 );
+}
+
+sub test_ie_mg {
+  my $ie = canon_ie_mg( shift );
+  return undef if length $ie != 13;
+  my @ie = split '', $ie;
+
+  my $c1 = - _dot_10( [1, 2, 1,  1, 2, 1, 2, 1, 2, 1, 2, undef, undef], \@ie ) % 10;
+  unless ( $ie[11] eq $c1 ) {
+    return 0;
+  }
+
+  my $s2 = _dot( [ 3, 2, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1 ], \@ie ) % 11;
+  return ( $s2==0 || $s2==1 && $ie[12]==0 ) ? 1 : 0;
+}
+
+sub format_ie_mg {
+  my $ie = canon_ie_mg shift;
+  $ie =~ s|^(...)(...)(...)(..)(..).*|$1.$2.$3/$4$5|; # 062.307.904/0081
+  return $ie;
+}
+
+sub parse_ie_mg {
+  my $ie = canon_ie_mg shift;
+  my ($municipio, $inscricao, $ordem, $dv) = $ie =~ /(\d{3})(\d{6})(\d{2})(\d{2})/;
+  if (wantarray) {
+    return ($municipio, $inscricao, $ordem, $dv);
+  }
+  return { 
+    municipio => $municipio, 
+    inscricao => $inscricao, 
+    ordem     => $ordem, 
+    dv        => $dv,
+  };
+}
+
+sub _dv_ie_mg {
+  my $base = shift; # expected to be canon'ed already ?!
+  my $valid = @_ ? shift : 1;
+  my $dev = $valid ? 0 : 2; # deviation (to make IE/MG invalid)
+  my @base = split '', substr($base, 0, 11);
+  my $dv1 = -_dot_10([ 1, 2, 1,  1, 2, 1, 2, 1, 2, 1, 2 ], \@base) % 10;
+  my $dv2 = (-_dot([ 3, 2, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ], [ @base, $dv1 ]) + $dev) % 11 % 10;
+  return ($dv1, $dv2) if wantarray;
+  substr($base, 11, 2) = "$dv1$dv2";
+  return $base;
+}
+sub random_ie_mg {
+  my $valid = @_ ? shift : 1; # valid IE/MG by default
+  my $base = sprintf "%011s", int(rand(1E11)); # 11 digits
+  return scalar _dv_ie_mg($base, $valid);
+}
+
+
+### PA ###
+
+sub canon_ie_pa {
+  return _canon_id(shift, size => 9);
+}
+
+### PB ###
+
+sub canon_ie_pb {
+  return _canon_id(shift, size => 9);
+}
 
 ### PR ###
 
@@ -195,7 +383,7 @@ sub parse_ie_ma {
 
 
 sub canon_ie_pr {
-  return _canon_i(shift, size => 10);
+  return _canon_id(shift, size => 10);
 }
 sub test_ie_pr {
   my $ie = canon_ie_pr shift;
@@ -211,7 +399,7 @@ sub test_ie_pr {
 }
 sub format_ie_pr {
   my $ie = canon_ie_pr shift;
-  $ie =~ s|^(...)(.....)(..).*|$1.$2-$4|;
+  $ie =~ s|^(...)(.....)(..).*|$1.$2-$3|;
   return $ie;
 }
 sub _dv_ie_pr {
@@ -239,12 +427,85 @@ sub parse_ie_pr {
   return { base => $base, dv => $dv };
 }
 
+### PE ###
+
+sub canon_ie_pe {
+  return _canon_id(shift, size => 14);
+}
+
+### PI ###
+
+sub canon_ie_pi {
+  return _canon_id(shift, size => 9);
+}
+
+### RJ ###
+
+sub canon_ie_rj {
+  return _canon_id(shift, size => 9);
+}
+
+### RN ###
+
+sub canon_ie_rn {
+  return _canon_id(shift, size => 9);
+}
+
+### RS ###
+
+sub canon_ie_rs {
+  return _canon_id(shift, size => 10);
+}
+
+### RO ###
+
+# http://www.sintegra.gov.br/Cad_Estados/cad_RO.html
+
+sub canon_ie_ro {
+  return _canon_id(shift, size => 14);
+}
+sub test_ie_ro {
+  my $ie = canon_ie_ro shift;
+  return undef if length $ie != 14;
+  my @ie = split '', $ie;
+  my $s1 = _dot([6, 5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2, 1], \@ie) % 11;
+  return $s1==0 || $ie[13]==0 && $s1==1;
+}
+sub format_ie_ro {
+  my $ie = canon_ie_ro shift;
+  $ie =~ s|^(.............)(.).*|$1-$2|;
+  return $ie;
+}
+sub _dv_ie_ro {
+	my $base = shift; # expected to be canon'ed already ?!
+	my $valid = @_ ? shift : 1;
+	my $dev = $valid ? 0 : 2; # deviation (to make IE-RO invalid)
+	my @base = split '', substr($base, 0, 13);
+	my $dv = (-_dot([6, 5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2], \@base)+$dev) % 11 % 10;
+	return ($dv) if wantarray;
+	substr($base, 13, 1) = $dv;
+	return $base;
+}
+sub random_ie_ro {
+	my $valid = @_ ? shift : 1; # valid IE-RO by default
+	my $base = sprintf "%013s", int(rand(1E13)); # 13 dgitos   # devia ter maior probabilidade para 000 00000 AAAAA D
+	return scalar _dv_ie_ro($base, $valid);
+}
+sub parse_ie_ro {
+  my $ie = canon_ie_ro shift;
+  my ($base, $dv) = $ie =~ /(\d{13})(\d{1})/;
+  if (wantarray) {
+    return ($base, $dv);
+  }
+  return { base => $base, dv => $dv };
+}
+
 ### RR ###
 
 # http://www.sintegra.gov.br/Cad_Estados/cad_RR.html
 
 sub canon_ie_rr {
-  return _canon_i(shift, size => 9);
+  return _canon_id(shift, size => 9);
 }
 sub test_ie_rr {
   my $ie = canon_ie_rr shift;
@@ -284,12 +545,16 @@ sub parse_ie_rr {
   return { base => $base, dv => $dv };
 }
 
+### SC ###
 
+sub canon_ie_sc {
+  return _canon_id(shift, size => 9);
+}
 
 ### SP ###
 
 sub canon_ie_sp {
-  return _canon_i(shift, size => 12);
+  return _canon_id(shift, size => 12);
 }   
 
 #SP - http://www.csharpbr.com.br/arquivos/csharp_mostra_materias.asp?escolha=0021
@@ -358,6 +623,17 @@ sub parse_ie_sp {
   return { base => $base, dv => $dv };
 }
 
+### SE ###
+
+sub canon_ie_se {
+  return _canon_id(shift, size => 9);
+}
+
+### TO ###
+
+sub canon_ie_to {
+  return _canon_id(shift, size => 11);
+}
 
 # a dispatch table is used here, because we know beforehand 
 # the list of Brazilian states. I am not sure it is
@@ -379,12 +655,30 @@ my %dispatch_table = (
   parse_ie_al => \&parse_ie_al,
 
   # AM
+  canon_ie_am => \&canon_ie_am, 
+
   # AP
+  test_ie_ap => \&test_ie_ap, 
+  canon_ie_ap => \&canon_ie_ap, 
+  format_ie_ap => \&format_ie_ap,
+  random_ie_ap => \&random_ie_ap,
+  parse_ie_ap => \&parse_ie_ap,
+
   # BA
+  canon_ie_ba => \&canon_ie_ba, 
+
   # CE
+  canon_ie_ce => \&canon_ie_ce, 
+
   # DF
+  canon_ie_df => \&canon_ie_df, 
+
   # ES
+  canon_ie_es => \&canon_ie_es, 
+
   # GO
+  canon_ie_go => \&canon_ie_go, 
+
   # MA
   test_ie_ma => \&test_ie_ma, 
   canon_ie_ma => \&canon_ie_ma, 
@@ -392,13 +686,30 @@ my %dispatch_table = (
   random_ie_ma => \&random_ie_ma,
   parse_ie_ma => \&parse_ie_ma,
 
-  #test_ie_mg => \&test_ie_mg,  # MG
+  # MG
+  test_ie_mg => \&test_ie_mg, 
+  canon_ie_mg => \&canon_ie_mg, 
+  format_ie_mg => \&format_ie_mg,
+  random_ie_mg => \&random_ie_mg,
+  parse_ie_mg => \&parse_ie_mg,
+
   # MT
+  canon_ie_mt => \&canon_ie_mt, 
+
   # MS
+  canon_ie_ms => \&canon_ie_ms, 
+
   # PE
+  canon_ie_pe => \&canon_ie_pe, 
+
   # PA
+  canon_ie_pa => \&canon_ie_pa, 
+
   # PB 
+  canon_ie_pb => \&canon_ie_pb, 
+
   # PI
+  canon_ie_pi => \&canon_ie_pi, 
 
   # PR
   test_ie_pr => \&test_ie_pr, 
@@ -408,8 +719,17 @@ my %dispatch_table = (
   parse_ie_pr => \&parse_ie_pr,
 
   # RJ
+  canon_ie_rj => \&canon_ie_rj, 
+
   # RN
+  canon_ie_rn => \&canon_ie_rn, 
+
   # RO
+  test_ie_ro => \&test_ie_ro, 
+  canon_ie_ro => \&canon_ie_ro, 
+  format_ie_ro => \&format_ie_ro,
+  random_ie_ro => \&random_ie_ro,
+  parse_ie_ro => \&parse_ie_ro,
   # RR
   test_ie_rr => \&test_ie_rr, 
   canon_ie_rr => \&canon_ie_rr, 
@@ -417,17 +737,19 @@ my %dispatch_table = (
   random_ie_rr => \&random_ie_rr,
   parse_ie_rr => \&parse_ie_rr,
   # RS
+  canon_ie_rs => \&canon_ie_rs, 
   # SC
+  canon_ie_sc => \&canon_ie_sc, 
   # SE
-
+  canon_ie_se => \&canon_ie_se, 
   # SP
   test_ie_sp => \&test_ie_sp, 
   canon_ie_sp => \&canon_ie_sp, 
   format_ie_sp => \&format_ie_sp,
   random_ie_sp => \&random_ie_sp,
   #parse_ie_sp
-
   # TO
+  canon_ie_to => \&canon_ie_to, 
 
 );
 
@@ -479,13 +801,15 @@ Business::BR::IE - Perl module to test for correct IE numbers
   test_ie('ac', '01.004.823/001-12') # 1
   test_ie('al', '24.000.004-8') # 1
   test_ie('ma', '12.000.038-5') # 1
+  test_ie('mg', '062.307.904/0081') #1
   test_ie('rr', '24.006.628-1') # 1
+  test_ie('ap', '03.012.345-9') # 1
 
 =head1 DESCRIPTION
 
 YET TO COME. Handles IE for the states of Acre (AC),
-Alagoas (AL), Maranho (MA), Paran (PR), Roraima (RR)
-and Sao Paulo (SP) by now.
+Alagoas (AL), Amap (AP), Maranho (MA), Minas Gerais (MG), Paran (PR), Rondnia (RO), 
+Roraima (RR) and Sao Paulo (SP) by now.
 
 =head2 EXPORT
 
@@ -623,6 +947,33 @@ and if it satisfies the check equation:
 
 =back
 
+=head2 MG
+
+The state of Minas Gerais uses:
+
+=over 4
+
+=item *
+
+13-digits number
+
+=item *
+
+the 11th and 12th are check digits
+
+=item *
+
+the usual formatting is like C<'062.307.904/0081'>
+
+=item *
+
+to determine if IE/MG number is correct, the computation rules in
+
+  http://www.sintegra.gov.br/Cad_Estados/cad_MG.html
+
+must be followed. (Yes, they are boring and hard to describe.)
+
+=back
 
 =head2 PR
 
@@ -644,6 +995,39 @@ the usual formatting is like C<'123.45678-50'>
 
 =back
 
+=head2 RO
+
+The state of Rondnia uses:
+
+=over 4
+
+=item *
+
+14-digits number
+
+=item *
+
+the 14th is a check digit
+
+=item *
+
+the usual formatting is like C<'0000000062521-3'>
+
+=item *
+
+if the IE-RO number is decomposed into digits like this
+
+  a_1 a_2 a_3 a_4 a_5 a_6 a_7 a_8 a_9 a_10 a_11 a_12 a_13 d_1 
+
+it is correct if it satisfies the check equation:
+
+  ( 6 a_1 + 5 a_2 + 4 a_3 + 3 a_4 + 2 a_5 +
+    9 a_6 + 8 a_7 + 7 a_8 + 6 a_9 + 5 a_10 +
+    4 a_11 + 3 a_12 + 2 a_13 + d_1            = 0  (mod 11) or
+                                              = 1  (mod 11) if d_1 = 0
+
+=back
+
 =head2 RR
 
 The state of Roraima uses:
@@ -712,7 +1096,7 @@ This documentation is faulty
 
 =item *
 
-If you want handling more than AC, AL, MA, PR and SP, you'll
+If you want handling more than AC, AL, MA, MG, PR, RO, RO and SP, you'll
 have to wait for the next releases
 
 =item *
@@ -742,11 +1126,10 @@ A. R. Ferreira, E<lt>ferreira@cpan.orgE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2005 by A. R. Ferreira
+Copyright (C) 2005-2007 by A. R. Ferreira
 
 This library is free software; you can redistribute it and/or modify
-it under the same terms as Perl itself, either Perl version 5.8.6 or,
-at your option, any later version of Perl 5 you may have available.
+it under the same terms as Perl itself.
 
 
 =cut
